#!/usr/bin/env python

from collections import defaultdict
from copy import copy
from sympy import *
from LMPS import ModularNumbers, LinearModularParametricSystem, vanishing_iter
from GAPI import GAPInterface
from MVS import Matrice, VectorSpace

# Legend:
# par:    parameter
# unk:    unknown
# cr:     canonical ring
# rel:    relation
# gen:    generator
# col:    column
# row:    row
# mod:    moduli space

class GodeauxAutomorphismComputer:
    def __init__(self, n, monomials, mod_pars, cr_gens, cr_rels, cr_rels_multipliers, sys_unks, sys_pars, sys_pars_coupling, alpha, perms, rho, psi):
        """
        monomials: list of monomials in cr_gens composing cr_rels
        Example: [x[1]**5, x[2]**5, ...]
        
        mod_pars: list of parameters of cr_rels
        Example: [b[1], b[2], ...]

        cr_gens: list of unknowns of cr_rels
        Example: [x[1], x[2], ...]

        cr_rels: list of polynomials
        Example: [x[1]**5 + b[1]*x[2]**5 + ..., ...]

        cr_rels_multipliers: list of exponents of xi that compares in
                             the equations after applying an
                             automorphism
        Example: [0]
        
        sys_unks: list of modular unknowns composing alpha
        Example: [i[2], i[3], ...]

        sys_pars: list of modular parameter which will come out in the
                  computation
        Example: [n[1,2], ...]
        
        sys_pars_coupling: dict of correspondance between mod_pars and
                           sys_pars
        Example: {(b[1], b[3]) -> n[1,2]+n[2,3], ...}

        alpha: dict of matrices representing an automorphism depending
               on a permutation
        Example: {1 -> Matrix(...), ...}

        perms: dict of permutation matrices
        Example: {1 -> Matrix(...), ...}
        """
        self.n = n
        self.csi = exp(Rational(2,n)*pi*I)
        self.monomials = monomials
        self.n_monomials = len(self.monomials)
        self.mod_pars = mod_pars
        self.n_mod_pars = len(self.mod_pars)
        self.cr_gens = cr_gens
        self.n_cr_gens = len(self.cr_gens)
        self.cr_rels = cr_rels
        self.n_cr_rels = len(self.cr_rels)
        self.cr_rels_multipliers = cr_rels_multipliers
        self.sys_unks = sys_unks
        self.n_sys_unks = len(self.sys_unks)
        self.sys_pars = sys_pars
        self.n_sys_pars = len(self.sys_pars)
        self.sys_pars_coupling = sys_pars_coupling
        self.alpha = alpha
        self.perms = perms
        self.rho = rho
        self.psi = psi
        self.n_perms = len(perms)
        self.gap = GAPInterface()

    def nullpars(self, f):
        """
        Given a list of 0es and 1s, returns the parameters in mod_pars
        corresponding to 0es.
        """
        return [self.mod_pars[t] for t in xrange(self.n_mod_pars) if f[t] == 0]

    def nonnullpars(self, f):
        """
        Given a list of 0es and 1s, returns the parameters in mod_pars
        corresponding to 1es.
        """
        return [self.mod_pars[t] for t in xrange(self.n_mod_pars) if f[t] != 0]

    def apply_linear_function(self, M, f, s):
        """
        Applies matrix M to expression f transforming the symbol in
        s[i] using the i-th row of M.
        """
        d = []
        g = f
        for x in s:
            d.append(Symbol("t", dummy = True))
        for i in xrange(len(s)):
            g = g.subs(s[i], d[i])
        for i in xrange(len(s)):
            tmp = M[i,:] * Matrice(s)
            g = g.subs(d[i], tmp[0,0])        
        return g

    def print_list_vs(self, l):
        tmp = [v.dimension for v in l]
        for y in set(tmp):
            print "        ", tmp.count(y), " of dimension ", y
            
    def compute_equations(self):
        """
        Compute the modular equations for every possible permutation
        and monomial.

        Output: self.equations

        Example: for n = 5, self.equations[4][(b_4, b_1)] =
                 (-n_{1,4} == i_2 + i_4 + 3*i_3)
        """
        p = [f.as_poly(*self.cr_gens) for f in self.cr_rels]

# q[h] is the list of equations in cr_rels transformed with the
# generic morphism with permutation h.
        self.transformed_cr_rels = {}
        q = {}
        for h in self.perms.keys():
            self.transformed_cr_rels[h] = [self.apply_linear_function(self.alpha*self.perms[h], f, self.cr_gens) for f in self.cr_rels]
            q[h] = [g.as_poly(*self.cr_gens) for g in self.transformed_cr_rels[h]]
            
        self.equations = defaultdict(dict)
        for h in self.perms.keys():
# We put all equation together (it is easy to see that there could not
# be mixing applying a morphism with permutation h).
            tmp_p = tmp_q = 0
            for i in xrange(len(p)):
                tmp_p += p[i]
                tmp_q += q[h][i]
                
            for term in self.monomials:
# After this cycle, c is the coefficient of term in the original
# equation, while d is the same in the transformed one.
                for index in xrange(len(p)):
# Polynomial divided by a term is the coefficient of the term.
                    c = (p[index] / term).as_poly(*self.mod_pars)
                    d = (q[h][index] / term).as_poly(*self.mod_pars)
                    if not c.is_zero: break
                assert not c.is_zero
# Do not need to check constant-coefficient terms.
                if c.is_constant: continue
                elif c != d:
# There's only one term: the coefficient.
                    c = list(c.iter_terms())[0]
                    d = list(d.iter_terms())[0]
                    i_c = list(c[1]).index(1)
                    i_d = list(d[1]).index(1)
# c and d are pairs (coefficient, (list of 1 an 0 corresponding to an
# element in mod_pars; i_c and i_d are the index of this element in
# mod_pars.
                    tmp = d[0]/c[0]
# Workaround to assist sympy: the case 4 is slightly different since
# sympy has an explicit expression for exp(pi/2).
                    if tmp.is_Pow and tmp.base == I:
                        tmp = tmp.exp
                    elif tmp.func == exp:
                        tmp = (log(tmp) * self.n/(2*pi*I)).expand()
# If the parameters are different, the equation is slightly harder
                    if c[1] != d[1]:
                        self.equations[h][(self.mod_pars[i_c],self.mod_pars[i_d])] = \
                            tmp.expand() + self.cr_rels_multipliers[index] - \
                            self.sys_pars_coupling[self.mod_pars[i_c], self.mod_pars[i_d]]
# If the parameters are equal, it is easier.
                    else:
                        self.equations[h][self.mod_pars[i_c]] = tmp.expand() + \
                            self.cr_rels_multipliers[index]
        
    def compute_solutions(self):
        """
        Compute the vector spaces that are locus of special
        automorphisms groups inside the moduli space.

        Output: self.automorphisms_gens

        Example: for n = 5, self.automorphisms_gens[origin] is a set
                 of generators of the automorphisms group of the
                 Fermat quintic.
        """
        self.syss = {}
        self.automorphisms_gens = defaultdict(set)
# We iterate through all possible vanishing of elements of mod_pars,
# that is, through all possible different systems obtained from the
# equality of the relations in cr_rels.
        for f in vanishing_iter([0,1], self.n_mod_pars):
            current_vector_spaces = set()
            nulls = self.nullpars(f)
            nonnulls = self.nonnullpars(f)
# Also, for the different permutations
            for h in self.perms.keys():
                sys_equations = []
# We put in sys_equations the equations of the systems relative to the
# current vanishing of mod_pars and permutation.
                for t in self.equations[h].keys():
# If t is just a parameter, we put the corresponding equation if it is
# in the nonnuls parameters.
                    if t in nonnulls:
                        sys_equations.append(self.equations[h][t])
                    elif t in nulls:
                        continue
# If t is a couple of parameter, we put the equation only if they are
# both in nonnuls, but if they are one in nonnulls and one in nulls,
# there is no solutions (to do this, we append the equation 1 == 0).
                    elif t[0] in nonnulls and t[1] in nonnulls:
                        sys_equations.append(self.equations[h][t])
                    elif t[0] in nulls and t[1] in nulls:
                        continue
                    else:
                        sys_equations.append(sympify(1))
                        
# Now we solve the parametric systems.
                sys_pars = [p for p in self.sys_pars \
                                if any([e.has(p) for e in sys_equations])]
                self.syss[f,h] = LinearModularParametricSystem(self.n, self.sys_unks, sys_pars, sys_equations)
                self.syss[f,h].solve()

# We compute the subspaces with special automorphisms groups. We put
# in gens_sample_solutions the matrices of automorphisms for the
# current vanishing of parameters and permutation, with some fixed
# value for the parameters (since the group does not depend on the
# particular value of the parameters).
                gens_sample_solutions = set()
                for s in self.syss[f,h].gens_sample_solutions():
                    gens_sample_solutions.add(Matrice((self.alpha*self.perms[h]).subs(s)))
                if gens_sample_solutions == set(): continue
# We construct the equations cutting the locus with the special
# automorphisms group inside the moduli space. There are the vanishing
# parameters equal to zero, but also the parameter whose ratio is
# forced to be a root of unity.
                for g in self.syss[f,h].iter_pars_solutions():
                    relations = nulls[:]
                    for p in self.syss[f,h].pars:
                        for x, y in self.sys_pars_coupling.iteritems():
                            if y == p:
                                break
                        assert y == p
                        relations.append(x[0] - self.csi**(p.subs(g)) * x[1])
# We put all vector subspaces with the corresponding automorphisms
# group in the dictionary automorphisms_gens, and add the various
# vector subspaces in the set current_vector_spaces.
                    v = VectorSpace(self.gap, self.mod_pars, relations)
                    self.automorphisms_gens[v].update(gens_sample_solutions)
                    current_vector_spaces.add(v)

# Finally, if two vector spaces are one contained in the other, we
# have to add the automorphisms of the bigger to the automorphisms of
# the smaller. This is needed only when the two vector spaces are in
# the same type of vanishing of the parameters, hence we can do the
# following computation one vanishing a time.
            for v in current_vector_spaces:
                for w in current_vector_spaces:
                    if v < w:
                        self.automorphisms_gens[v].update(self.automorphisms_gens[w])
            
    def regroup_solutions(self):
        """
        For every possible set of generators of some automorphisms
        group, compute GAP's id of the group. Build a dictionary,
        spaces, indexed by GAP's ids and with values lists of vector
        spaces. In the end, remove from these lists the irrelevant
        spaces (that is, the spaces contained in another space with
        the same automorphisms group).

        Output: self.spaces

        Example: for n = 5, self.spaces of the group with 100 elements
                 is the list with just the origin.
        """
        group_id = {}
        spaces = defaultdict(set)
        
        def compare_matrices(x, y):
            """
            Gives a unique sorting criterion for matrices.
            """
            if x.shape[0] != y.shape[0]: return cmp(x.shape[0], y.shape[0])
            if x.shape[1] != y.shape[1]: return cmp(x.shape[1], y.shape[1])
            for i in xrange(x.shape[0]):
                for j in xrange(x.shape[1]):
                    if x[i,j] != y[i,j]: return cmp(x[i,j], y[i,j])
            return 0

# We build a dictionary, spaces, indexed by the GAP's group ids (note
# that we compute an id only once for equal sets of generators) of the
# automorphisms groups generated by keys of automorphisms_gens. The
# values are lists of vector spaces with the same automorphisms group.
        for v in self.automorphisms_gens.keys():
            gens = list(self.automorphisms_gens[v])
            gens.sort(compare_matrices)
            gens = tuple(gens)
            if gens not in group_id.keys():
                group_id[gens] = self.gap.IdSmallGroup(gens,
                                                       self.psi,
                                                       self.rho)
            spaces[group_id[gens]].add(v)
        print "Set of generators: ", len(group_id)
        print "Groups: ", len(spaces)
        print "Vector spaces (start): ", Add(*[len(q) for q in spaces.values()])

# We remove, in the values of spaces, subvectorspaces with the same
# automorphisms group.
        for x in spaces.keys():
            l = list(spaces[x])
            N = len(l)
            remove = [False] * N
            for v in xrange(N):
                if remove[v]: continue
                for w in xrange(v+1, N):
                    if remove[w]: continue
                    if l[v] < l[w]:
                        remove[v] = True
                        break
                    elif l[v] > l[w]:
                        remove[w] = True
            spaces[x] = [l[i] for i in xrange(N) if not remove[i]]
        print "Vector spaces (end): ", \
            Add(*[len(q) for q in spaces.values()])
        print

# Printing
        for x in spaces.keys():
            print x
            for v in spaces[x]:
                print v
            print

        print "Resume:"
        print "    groups: ", len(spaces)

        tmp = spaces.keys()
        tmp.sort()

        for x in tmp:
            print "    group ", x, " with ", len(spaces[x]), "components, of which"
            self.print_list_vs(spaces[x])
            
    def compute(self):
        self.compute_equations()
        self.compute_solutions()
        self.regroup_solutions()

################################################################################
################################################################################
################################################################################

def z5():
# Generating the needed symbols for the case Z_5
    x = {}
    b = {}
    c = {}
    i = {}
    for t in [1,2,3,4]:
        x[t] = Symbol('x_%d' % t)
        b[t] = Symbol('b_%d' % t)
        c[t] = Symbol('c_%d' % t)
        i[t] = Symbol('i_%d' % t, integer = True)

# Generators of the canonical ring
    cr_gens = [x[1], x[2], x[3], x[4]]
# Parameters of the moduli space
    mod_pars = [b[1], b[2], b[3], b[4], c[1], c[2], c[3], c[4]]
    sys_unks = [i[2], i[3], i[4]]
    
# Generating monomials for the equation of the quintic
    monomials = [x[1] ** 5, x[2] ** 5, x[3] ** 5, x[4] ** 5,\
                 x[2] * x[3] ** 3 * x[4], x[1] ** 3 * x[3] * x[4], x[1] * x[2] * x[4] ** 3, x[1] * x[2] ** 3 * x[3],\
                 x[2] ** 2 * x[3] * x[4] ** 2, x[1] * x[3] ** 2 * x[4] ** 2, x[1] ** 2 * x[2] ** 2 * x[4], x[1] ** 2 * x[2] * x[3] ** 2]

# Relations in the canonical ring
    cr_rels = [x[1] ** 5 + x[2] ** 5 + x[3] ** 5 + x[4] ** 5 +\
               b[1] * x[2] * x[3] ** 3 * x[4] + b[2] * x[1] ** 3 * x[3] * x[4] + b[3] * x[1] * x[2] * x[4] ** 3 + b[4] * x[1] * x[2] ** 3 * x[3] +\
               c[1] * x[2] ** 2 * x[3] * x[4] ** 2 + c[2] * x[1] * x[3] ** 2 * x[4] ** 2 + c[3] * x[1] ** 2 * x[2] ** 2 * x[4] + c[4] * x[1] ** 2 * x[2] * x[3] ** 2]
    cr_rels_poly = [f.as_poly(*cr_gens) for f in cr_rels]

    r = exp(Rational(2,5)*pi*I)

##########
# perms[t] = permutation matrices of type t
##########
    perms = {}
    perms[1] = Matrice(eye(4))
    perms[2] = Matrice(zero(4))
    perms[2][0,1] = 1
    perms[2][1,3] = 1
    perms[2][2,0] = 1
    perms[2][3,2] = 1
    perms[4] = perms[2]**2
    perms[3] = perms[2] * perms[4]

##########
# alpha = generic automorphism of type 1
##########
    alpha = Matrice(zero(4))
    alpha[0,0] = 1
    for t in xrange(1,4):
        alpha[t,t] = r ** i[t+1]

##########
# psi = generator of the normal subgroup to get the projective matrix
# group
# rho = generator of the normal subgroup that gives the action on the
# canonical ring
##########
    psi = Matrice(eye(4) * r)
    rho = Matrice(eye(4))
    for t in xrange(1,4):
        rho[t,t] = r ** t

##########
# sys_pars_coupling[b[1], b[2]] = n[1,2]
##########
    sys_pars_coupling = {}
    sys_pars = []
    for t in xrange(4):
        for s in xrange(t+1, 4):
            sys_pars_coupling[mod_pars[t], mod_pars[s]] = Symbol("n_{%d,%d}" % (t+1, s+1), integer = True)
            sys_pars_coupling[mod_pars[t+4], mod_pars[s+4]] = Symbol("m_{%d,%d}" % (t+1, s+1), integer = True)
            sys_pars_coupling[mod_pars[s], mod_pars[t]] = -sys_pars_coupling[mod_pars[t], mod_pars[s]]
            sys_pars_coupling[mod_pars[s+4], mod_pars[t+4]] = -sys_pars_coupling[mod_pars[t+4], mod_pars[s+4]]
            sys_pars.append(sys_pars_coupling[mod_pars[t], mod_pars[s]])
            sys_pars.append(sys_pars_coupling[mod_pars[t+4], mod_pars[s+4]])

    cr_rels_multipliers = [0]
    G = GodeauxAutomorphismComputer(5, monomials, mod_pars, cr_gens, cr_rels, cr_rels_multipliers, sys_unks, sys_pars, sys_pars_coupling, alpha, perms, rho, psi)
    G.compute()
    
def z4():
# Generating the needed symbols for the case Z_4
    x = {}
    y = {}
    a = {}
    b = {}
    c = {}
    i = {}
    j = {}
    for t in [1,2,3]:
        for s in xrange(t, 4):
            x[t,s] = Symbol('x_%d_%d' % (t,s))
    y[1] = Symbol('y_1')
    y[3] = Symbol('y_3')
    a[1,3] = Symbol('a_1_3')
    a[1,2,3] = Symbol('a_1_2_3')
    b[1,2] = Symbol('b_1_2')
    b[2,3] = Symbol('b_2_3')
    c[1,3] = Symbol('c_1_3')
    c[3,1] = Symbol('c_3_1')
    c[1,2] = Symbol('c_1_2')
    c[2,3] = Symbol('c_2_3')
    i[1] = Symbol('i_1', integer = True)
    i[3] = Symbol('i_3', integer = True)
    j[1] = Symbol('j_1', integer = True)

# Generators of the canonical ring
    cr_gens = [x[1,1], x[2,2], x[3,3], x[2,3], x[1,3], x[1,2], y[1], y[3]]
# Parameters of the moduli space
    mod_pars = [a[1,3], a[1,2,3], b[1,2], b[2,3], c[1,3], c[3,1], c[1,2], c[2,3]]
    sys_unks = [i[1], i[3], j[1]]
    
# Generating monomials for the canonical relations
    monomials = [x[1,1]**2, x[2,2]**2, x[3,3]**2, x[1,3]**2, x[1,3]*x[2,2], y[1]*y[3], y[1]*x[1,2], y[3]*x[2,3],
                 x[1,1]*x[1,3], x[1,3]*x[3,3], x[1,2]**2, x[2,3]**2, y[1]**2, y[3]**2]

# Relations in the canonical ring
    cr_rels = [x[1,1]**2 + x[2,2]**2 + x[3,3]**2 + a[1,3]*x[1,3]**2 + a[1,2,3]*x[1,3]*x[2,2] + y[1]*y[3] + b[1,2]*y[1]*x[1,2] + b[2,3]*y[3]*x[2,3],
               c[1,3]*x[1,1]*x[1,3] + c[3,1]*x[1,3]*x[3,3] + c[1,2]*x[1,2]**2 + c[2,3]*x[2,3]**2 + y[1]**2 + y[3]**2]
    cr_rels_poly = [f.as_poly(*cr_gens) for f in cr_rels]

    r = exp(2*pi*I/4)

##########
# perms[t] = permutation matrices of type t
##########
    perms = {}
    perms[1] = Matrice(eye(8))
    perms[3] = Matrice(zero(8))
    perms[3][0,2] = 1
    perms[3][2,0] = 1
    perms[3][1,1] = 1
    perms[3][3,5] = 1
    perms[3][5,3] = 1
    perms[3][4,4] = 1
    perms[3][6,7] = 1
    perms[3][7,6] = 1

##########
# alpha = generic automorphism of type 1
##########
    alpha = Matrice(eye(8))
    alpha[0,0] = r**(2*i[1])
    alpha[2,2] = r**(2*i[3])
    alpha[3,3] = r**(i[3])
    alpha[4,4] = r**(i[1]+i[3])
    alpha[5,5] = r**(i[1])
    alpha[6,6] = r**(j[1])
    alpha[7,7] = r**(-j[1])

##########
# psi = generator of the normal subgroup to get the projective matrix
# group
# rho = generator of the normal subgroup that gives the action on the
# canonical ring
##########
    psi = Matrice(eye(8) * (-1)) # don't really needed for Z_4
    rho = Matrice(eye(8))
    rho[0,0] = r**2
    rho[2,2] = r**2
    rho[3,3] = r
    rho[5,5] = r**3
    rho[6,6] = r
    rho[7,7] = r**3
    
##########
# sys_pars_coupling[b[1], b[2]] = n[1,2]
##########
    sys_pars_coupling = {}
    sys_pars = []
    sys_pars_coupling[b[1,2], b[2,3]] = Symbol("n_1", integer = True)
    sys_pars_coupling[c[1,3], c[3,1]] = Symbol("n_2", integer = True)
    sys_pars_coupling[c[1,2], c[2,3]] = Symbol("n_3", integer = True)
    sys_pars_coupling[b[2,3], b[1,2]] = -Symbol("n_1", integer = True)
    sys_pars_coupling[c[3,1], c[1,3]] = -Symbol("n_2", integer = True)
    sys_pars_coupling[c[2,3], c[1,2]] = -Symbol("n_3", integer = True)
    sys_pars.append(sys_pars_coupling[b[1,2], b[2,3]])
    sys_pars.append(sys_pars_coupling[c[1,3], c[3,1]])
    sys_pars.append(sys_pars_coupling[c[1,2], c[2,3]])

    cr_rels_multipliers = [0, 2*j[1]]
    G = GodeauxAutomorphismComputer(4, monomials, mod_pars, cr_gens, cr_rels, cr_rels_multipliers, sys_unks, sys_pars, sys_pars_coupling, alpha, perms, rho, psi)
    G.compute()

def z3():
# Generating the needed symbols for the case Z_3
    x = {}
    y = {}
    z = {}
    a = {}
    b = {}
    c = {}
    d = {}
    i = {}
    j = {}
    x[1,1] = Symbol('x_1_1')
    x[1,2] = Symbol('x_1_2')
    x[2,2] = Symbol('x_2_2')
    y[0] = Symbol('y_0')
    y[1] = Symbol('y_1')
    y[2] = Symbol('y_2')
    z[1] = Symbol('z_1')
    z[2] = Symbol('z_2')
    a[1] = Symbol('a_1')
    a[2] = Symbol('a_2')
    b[1] = Symbol('b_1_1')
    b[1,2] = Symbol('b_1_2')
    b[2] = Symbol('b_2_2')
    c[1] = Symbol('c_1')
    c[2] = Symbol('c_2')
    d[1] = Symbol('d_1')
    d[2] = Symbol('d_2')
    i[1] = Symbol('i_1', integer = True)
    i[3] = Symbol('i_3', integer = True)
    j[1] = Symbol('j_1', integer = True)
    k = Symbol('k', integer = True)

    q_0 = x[1,2]*(y[0]**2 - y[1]*y[2]) - x[1,1]*(y[2]**2 - y[0]*y[1]) - x[2,2]*(y[1]**2 - y[0]*y[2]) + \
        a[1]*x[1,1]*x[1,2]*y[1] + a[2]*x[1,2]*x[2,2]*y[2] - b[1]*x[1,1]**3 + b[1,2]*x[1,2]**3 - b[2]*x[2,2]**3
    p_0 = y[0]**3 + y[1]**3 + y[2]**3 - 3*y[0]*y[1]*y[2] +\
        a[1]*x[1,1]*y[0]*y[1] + a[2]*x[2,2]*y[0]*y[2] - (a[1]+a[2])*x[1,2]*y[1]*y[2] +\
        a[1]*x[2,2]*y[1]**2 + a[2]*x[1,1]*y[2]**2 + (b[1]+b[1,2]+b[2])*x[1,2]**2*y[0] +\
        b[2]*x[2,2]**2*y[1] - (b[1]+b[1,2])*x[1,1]*x[1,2]*y[1] +\
        b[1]*x[1,1]**2*y[2] - (b[1,2]+b[2])*x[1,2]*x[2,2]*y[2] +\
        c[1]*x[1,1]**3 + c[2]*x[2,2]**3 + (c[1]+c[2])*x[1,2]**3 +\
        d[1]*x[1,1]**2*y[2] + d[2]*x[2,2]**2*y[1] +\
        d[1]*x[1,2]*x[2,2]*y[2] + d[2]*x[1,1]*x[1,2]*y[1]
    
# Generators of the canonical ring
    cr_gens = [x[1,1], x[1,2], x[2,2], y[0], y[1], y[2]]
# Parameters of the moduli space
    mod_pars = [a[1], a[2], b[1], b[1,2], b[2], c[1], c[2], d[1], d[2]]
    sys_unks = [k]
    
# Generating monomials for the canonical relations
    monomials = [x[1,2]*y[0]**2, x[1,2]*y[1]*y[2], x[1,1]*y[2]**2, x[1,1]*y[0]*y[1], x[2,2]*y[1]**2, x[2,2]*y[0]*y[2], \
        x[1,1]*x[1,2]*y[1], x[1,2]*x[2,2]*y[2], x[1,1]**3, x[1,2]**3, x[2,2]**3, \
        y[0]**3, y[1]**3, y[2]**3, y[0]*y[1]*y[2], \
        x[1,1]*y[0]*y[1], x[2,2]*y[0]*y[2], x[1,2]*y[1]*y[2], \
        x[2,2]*y[1]**2, x[1,1]*y[2]**2, x[1,2]**2*y[0], \
        x[2,2]**2*y[1], x[1,1]*x[1,2]*y[1], \
        x[1,1]**2*y[2], x[1,2]*x[2,2]*y[2], \
        x[1,1]**3, x[2,2]**3, x[1,2]**3, \
        x[1,1]**2*y[2], x[2,2]**2*y[1], \
        x[1,2]*x[2,2]*y[2], x[1,1]*x[1,2]*y[1]]

# Relations in the canonical ring
    cr_rels = [q_0, p_0]
    cr_rels_poly = [f.as_poly(*cr_gens) for f in cr_rels]

    r = exp(2*pi*I/3)

##########
# perms[t] = permutation matrices of type t
##########
    perms = {}
    perms[1] = Matrice(eye(6))
    perms[2] = Matrice(zero(6))
    perms[2][0,2] = 1
    perms[2][2,0] = 1
    perms[2][1,1] = 1
    perms[2][3,3] = 1
    perms[2][4,5] = 1
    perms[2][5,4] = 1

##########
# alpha = generic automorphism of type 1
##########
    alpha = Matrice(eye(6))
    alpha[1,1] = r**(k)
    alpha[2,2] = r**(2*k)
    alpha[3,3] = r**(k)
    alpha[4,4] = r**(2*k)

##########
# psi = generator of the normal subgroup to get the projective matrix
# group
# rho = generator of the normal subgroup that gives the action on the
# canonical ring
##########
    psi = Matrice(eye(6) * r) # don't really needed for Z_4
    rho = Matrice(eye(6))
    rho[1,1] = r
    rho[2,2] = r**2
    rho[3,3] = r
    rho[4,4] = r**2
    
##########
# sys_pars_coupling[b[1], b[2,2]] = n[1,2]
##########
    sys_pars_coupling = {}
    sys_pars = []
    sys_pars_coupling[a[1], a[2]] = Symbol("n_1", integer = True)
    sys_pars_coupling[b[1], b[2]] = Symbol("n_2", integer = True)
    sys_pars_coupling[c[1], c[2]] = Symbol("n_3", integer = True)
    sys_pars_coupling[d[1], d[2]] = Symbol("n_4", integer = True)
    sys_pars_coupling[a[2], a[1]] = -Symbol("n_1", integer = True)
    sys_pars_coupling[b[2], b[1]] = -Symbol("n_2", integer = True)
    sys_pars_coupling[c[2], c[1]] = -Symbol("n_3", integer = True)
    sys_pars_coupling[d[2], d[1]] = -Symbol("n_4", integer = True)
    sys_pars.append(sys_pars_coupling[a[1], a[2]])
    sys_pars.append(sys_pars_coupling[b[1], b[2]])
    sys_pars.append(sys_pars_coupling[c[1], c[2]])
    sys_pars.append(sys_pars_coupling[d[1], d[2]])

    cr_rels_multipliers = [0, 0]
    G = GodeauxAutomorphismComputer(3, monomials, mod_pars, cr_gens, cr_rels, cr_rels_multipliers, sys_unks, sys_pars, sys_pars_coupling, alpha, perms, rho, psi)
    G.compute()

if __name__ == "__main__":
    i = sys.argv[1]
    if i == "3":
        z3()
    elif i == "4":
        z4()
    elif i == "5":
        z5()
    else:
        print "Usage:\n\t" + sys.argv[0] + " (3|4|5)"

