#!/usr/bin/env python

from sympy import *

def vanishing_iter(values, n):
    """
    Generates all lists of 0 and 1 of length n.
    """
    def inc(state, values):
        if len(state) == 0: return ()
        index = values.index(state[-1])
        if index+1 < len(values):
            return state[:-1] + (values[index+1],)
        else:
            return inc(state[:-1], values) + (values[0],)

    state = (values[0],) * n
    yield state
    while state != (values[-1],) * n:
        state = inc(state, values)
        yield state

class ModularNumbers:
    def __init__(self, n):
        self.n = n
        self.multiplied_by = [-1] * self.n
        self.representant_of = [-1] * self.n
        self.difficulty_of = [0] * self.n
        for i in xrange(1,self.n):
            g = self.simple_gcd(i, self.n)
            self.representant_of[i] = g
            if self.difficulty_of[g] != 0: self.difficulty_of[i] = self.difficulty_of[g]
            else: self.difficulty_of[i] = len(set([(k * i) % self.n for k in xrange(self.n)]))
            self.multiplied_by[i] = self.simple_inverse(n/g, i/g)

        self.multiplied = [0] * self.n
        for i in xrange(self.n):
            self.multiplied[i] = [0] * self.n
            for j in xrange(self.n):
                try:
                    index = [(k * i) % self.n for k in xrange(self.n)].index(j)
                    self.multiplied[i][j] = index
                except:
                    pass

        self.representant_of[0] = 0
        self.difficulty_of[0] = 0
        self.multiplied_by[0] = 0

    def simple_gcd(self, a, b):
        if a == 0: return b
        elif b == 0: return a
        else: return self.simple_gcd(b % a, a)

    def simple_inverse(self, n, x):
        return [(x * i) % n for i in xrange(n)].index(1)
        
    def gcd(self, a, b):
        a = int(a)
        b = int(b)
        if a == 0: return self.representant_of[b], [0, self.multiplied_by[b]]
        r = [a,b]
        x = [1,0]
        y = [0,1]
        while r[-1] != 0:
            x.append(x[-2] - r[-2] / r[-1] * x[-1])
            y.append(y[-2] - r[-2] / r[-1] * y[-1])
            r.append(a*x[-1] + b*y[-1])
        return self.representant_of[r[-2]], [(self.multiplied_by[r[-2]] * x[-2]) % self.n, (self.multiplied_by[r[-2]] * y[-2]) % self.n]
        
    def polygcd(self, l):
        for i in xrange(len(l)):
            l[i] = int(l[i])
        if len(l) == 2:
            return self.gcd(*l)
        coefficient = [1] * len(l)
        g = l[0]
        for i in xrange(1, len(l)):
            h, [u, v] = self.gcd(g, l[i])
            for j in xrange(i):
                coefficient[j] = coefficient[j] * u
            coefficient[i] = v
            g = h
        return self.representant_of[g], [(c * self.multiplied_by[g]) % self.n for c in coefficient]

class LinearModularParametricSystem(ModularNumbers):
    def __init__(self, n, unks, pars, equations):
        ModularNumbers.__init__(self, n)        
        self.unks = unks
        self.n_unks = len(self.unks)
        self.pars = pars
        self.n_pars = len(self.pars)
        self.equations = [self.normalize_equation(e.as_poly(*(self.unks + self.pars))) for e in equations]
        self.n_equations = len(self.equations)
        self.solved = False
        self.rels = {}
        self.grps = {}
        self.pars_rels = {}
        self.pars_grps = {}
        if self.equations == []:
            self.solved = True
            for x in self.unks:
                self.grps[x] = self.n
            for x in self.pars:
                self.pars_grps[x] = self.n

    def normalize_equation(self, e):
        new_e = 0
        for c, m in e.iter_terms():
            new_t = (int(c) % self.n)
            for i in xrange(len(e.symbols)):
                new_t *= e.symbols[i] ** m[i]
            new_e += new_t
        return new_e.as_poly(*e.symbols)
            
    def coefficient(self, eq, symbol):
        m = [0] * (self.n_unks + self.n_pars)
        try:
            m[self.unks.index(symbol)] = 1
        except:
            try:
                m[self.pars.index(symbol)+self.n_unks] = 1
            except:
                pass
        return int(eq.coeff(*m))

    def insert_linear_combination_of_equations(self, index, coefficients):
        e = Add(*[coefficients[i] * self.equations[i].as_basic() for i in xrange(self.n_equations)]).as_poly(*(self.unks + self.pars))
        self.equations.insert(0, self.normalize_equation(e))
        self.n_equations += 1

    def subst_with_linear_combination(self, index, coefficients):
        e = Add(*[coefficients[i] * self.equations[i].as_basic() for i in xrange(self.n_equations)]).as_poly(*(self.unks + self.pars))
        self.equations[index] = self.normalize_equation(e)

    def remove_equation(self, index):
        del(self.equations[index])
        self.n_equations -= 1
        
    def solve(self):
        if self.solved: return

        if 1 in self.equations:
            self.solved = True
            self.rels = {}
            self.grps = {}
            self.pars_rels = {}
            self.pars_grps = {}
            return            
        
        STAGE_UNKS, STAGE_PARS = (0,1)
        for stage, unks, rels, grps in ((STAGE_UNKS, self.unks, self.rels, self.grps), (STAGE_PARS, self.pars, self.pars_rels, self.pars_grps)):
            if unks != []:
                if stage == STAGE_PARS:
                    for e in self.rels.values():
                        f = self.normalize_equation((2*e.as_basic()).as_poly(*(self.unks + self.pars)))
                        if not any([f.has(unk) for unk in self.unks]) and any([f.has(par) for par in self.pars]):
                            self.equations.append(f)
                            self.n_equations += 1
                while self.n_equations > 0:
                    coeff_unks = [[self.coefficient(e, u) for e in self.equations] for u in unks]
                    difficulties = [self.difficulty_of[self.polygcd(c)[0]] for c in coeff_unks]
                    index = difficulties.index(max(difficulties))
                    if difficulties[index] == 0: # end of unks determination
                        break
                    unk = unks[index]
                    multiplications = self.polygcd(coeff_unks[index])[1]
                    self.insert_linear_combination_of_equations(0, multiplications)
                    for i in xrange(1, self.n_equations):
                        tmp = self.polygcd([self.coefficient(self.equations[i], unk), self.coefficient(self.equations[0], unk)])[0]
                        self.subst_with_linear_combination(i, [-self.coefficient(self.equations[i], unk) / tmp] + ([0] * (i-1)) + [self.coefficient(self.equations[0], unk) / tmp] + ([0] * (self.n_equations - i - 1)))
                    if self.coefficient(self.equations[0], unk) != 1:
                        grps[unk] = self.coefficient(self.equations[0], unk)
                    for g in rels.keys():
                        multiplication = self.multiplied[self.coefficient(self.equations[0], unk)][self.coefficient(rels[g], unk)]
                        if int(multiplication * self.coefficient(self.equations[0], unk) - self.coefficient(rels[g], unk)) % self.n == 0:
                            rels[g] = self.normalize_equation((rels[g].as_basic() - self.equations[0].as_basic() * multiplication).as_poly(*(self.unks+self.pars)))
                    if stage == STAGE_PARS:
                        for g in self.rels.keys(): # normalization of previously found relations
                            multiplication = self.multiplied[self.coefficient(self.equations[0], unk)][self.coefficient(self.rels[g], unk)]
                            if int(multiplication * self.coefficient(self.equations[0], unk) - self.coefficient(self.rels[g], unk)) % self.n == 0:
                                self.rels[g] = self.normalize_equation((self.rels[g].as_basic() - self.equations[0].as_basic() * multiplication).as_poly(*(self.unks+self.pars)))
                    rels[unk] = self.equations[0]
                    self.remove_equation(0)
                for unk in unks: # adding last generators which doesn't have to satisfies relations
                    if unk not in rels.keys():
                        grps[unk] = self.n
        if set([x.as_basic() for x in self.equations]) not in [set([]), set([sympify(0)]), set([0])]: # 1 = 0
            self.rels = {}
            self.grps = {}
            self.pars_rels = {}
            self.pars_grps = {}
        self.solved = True

    def iter_solutions(self):
        """
        Returns a list of dictionary similar to: {i : 1, j : 3 + n}, where i and
        j are the unknowns of the system and n is a parameter. If the system is
        not solved, this returns nothing.
        """
        if not self.solved: return
        for g in vanishing_iter(range(self.n), len(self.grps)):
            gens_values = {}
            other_values = {}
            for i, (gen, grp) in enumerate(self.grps.iteritems()):
                gens_values[gen] = g[i]
            ok = True
            for unk, e in self.rels.iteritems():
                f = e.subs(gens_values)
                if unk in self.grps.keys():
                    if int(f) % self.n != 0:
                        ok = False
                        break
                else:                    
                    other_values[unk] = self.normalize_equation(solve(f, unk)[0].as_poly(*(self.unks+self.pars))).as_basic()
            if ok:
                gens_values.update(other_values)
                yield gens_values

    def iter_pars_solutions(self):
        """
        Returns a list of dictionary similar to: {n : 1, m : 3}, where n and
        m are the parameters of the system. If the system is not solved, this
        returns nothing.
        """
        if not self.solved: return
        for g in vanishing_iter(range(self.n), len(self.pars_grps)):
            gens_values = {}
            other_values = {}
            ok = True
            for i, (gen, grp) in enumerate(self.pars_grps.iteritems()):
                if int(grp * g[i]) % self.n != 0:
                    ok = False
                    break
                gens_values[gen] = g[i]
            if not ok: continue
            for unk, e in self.pars_rels.iteritems():
                if unk not in self.pars_grps.keys():
                    f = e.subs(gens_values).as_basic()
                    other_values[unk] = self.normalize_equation(solve(f, unk)[0].as_poly(*(self.pars))).as_basic()
            gens_values.update(other_values)
            yield gens_values

    def gens_sample_solutions(self):
        """
        Returns a list of dictionary similar to: {i : 1, j : 3}, where i and
        j are the unknowns of the system and the parameters are set to 0. If the
        system is not solved, this returns nothing.
        """
        if not self.solved: return
        gens_values = {}
        gens = []
        halfgens = []
        for gen in self.grps.keys():
            if gen not in self.rels.keys():
                gens_values[gen] = 0
                gens.append(gen)
            else:
                halfgens.append(gen)

        pars_values = {}
        for par in self.pars:
            pars_values[par] = 0

        other_values = {}
        other_values.update(pars_values)
        for gen in gens:
            gens_values[gen] = 1
            tmp_eqs = []
            for unk in halfgens:
                f = self.rels[unk].subs(gens_values).subs(pars_values)
                tmp_eqs.append(f)
            lmps = LinearModularParametricSystem(self.n, halfgens, [], tmp_eqs)
            lmps.solve()
            for g in lmps.iter_solutions():
                other_values = {}
                other_values.update(gens_values)
                other_values.update(pars_values)
                other_values.update(g)                    
                for unk, e in self.rels.iteritems():
                    if unk not in halfgens:
                        f = e.subs(other_values).as_basic()
                        tmp = solve(f, unk)[0].as_poly(*(self.unks+self.pars))
                        other_values[unk] = self.normalize_equation(tmp).as_basic()
                yield other_values
            gens_values[gen] = 0
   
    def __str__(self):
        s = ""
        if self.solved:
            s += "Unks:\n - gens: {"
            for u in self.grps.keys(): s += u.__str__() + ", "
            if self.grps == {}: s += ", "
            s = s[:-2] + "}\n - rels: {"
            for r in self.rels: s += r.as_basic().__str__() + ", "
            if self.rels == {}: s += ", "
            s = s[:-2] + "}\n - grps: "
            for i in self.grps.values(): s += "Z_" + i.__str__() + " x "
            if self.grps == {}: s += "{} x "
            s = s[:-3] + "\n"
            s += "Pars:\n - gens: {"
            for p in self.pars_grps.keys(): s += p.__str__() + ", "
            if self.pars_grps == {}: s += ", "
            s = s[:-2] + "}\n - rels: {"
            for r in self.pars_rels: s += r.as_basic().__str__() + ", "
            if self.pars_rels == {}: s += ", "
            s = s[:-2] + "}\n - grps: "
            for i in self.pars_grps.values(): s += "Z_" + i.__str__() + " x "
            if self.pars_grps == {}: s += "{} x "
            s = s[:-3] + ""
        else:
            s += "Not solved; equations:\n"
            for e in self.equations:
                s += e.as_basic().__str__() + "\n"
        return s
