#!/usr/bin/env python

from sympy import *

class Matrice(Matrix):
    """
    We would like an immutable matrix, so we can use it as a dictionary key and as an element of a set. So we do this, provided that changing a matrix in a set or in a dictionary will screw up everything.
    """
    def __hash__(self):
        return self.hash()

class VectorSpace:
    """
    Represent a vector subspace of the vector space <gens>, defined by
    linear relations rels. Binary relations represent containment.
    """
    def __init__(self, gap, gens, rels = []):
        self.gap = gap
        self.gens = gens
        self.n_gens = len(self.gens)
        if rels != []:
            self.M = Matrice(zeros((len(rels), self.n_gens)))
            for i, r in enumerate(rels):
                r = r.as_poly(*gens)
                for coeff, monom in r.iter_terms():
                    try:
                        j = list(monom).index(1)
                        self.M[i, j] = coeff
                    except ValueError:
                        pass
        else:
            self.M = zeros((1, self.n_gens))
        self.basis = self.gap.NullspaceMat(Matrice(self.M.transpose()))
        self.dimension = len(self.basis)
    
    def __le__(self, other):
        if self.gens != other.gens or self.dimension > other.dimension:
            return False
        for x in self.basis:
            if simplify((other.M * x).norm()) != 0:
                return False
        return True
    
    def __lt__(self, other):
        if self.gens != other.gens or self.dimension >= other.dimension: return False
        for x in self.basis:
            if simplify((other.M * x).norm()) != 0:
                return False
        return True
    
    def __eq__(self, other):
        if (other.gens, other.dimension) != (self.gens, self.dimension):
            return False
        return self <= other and other <= self
    
    def __ne__(self, other):
        return not (self == other)
    
    def __hash__(self):
        h = 0
        for x in self.gens + self.basis:
            h += x.__hash__()
        return h
    
    def __str__(self):
        s = "Vector space of dimension " + str(self.dimension)
        s += " with basis: "
        for b in self.basis:
            s += b.transpose().__str__() + " "
        return s

